import argparse
import os
import sys
import pkg_resources

import dbt.exceptions
import dbt.ui

from dbt.logger import log_manager, GLOBAL_LOGGER as logger
from dbt.config.profile import DEFAULT_PROFILES_DIR

from fal.run_scripts import run_global_scripts, run_scripts
from fal.dag import FalScript, ScriptGraph
from fal.utils import print_run_info
from faldbt.lib import DBT_VCURRENT, DBT_V1
from faldbt.project import FalDbt, FalGeneralException, FalProject


def run_fal(argv):

    # fmt: off
    parser = argparse.ArgumentParser(
        description="Run Python scripts on dbt models",
        usage="""fal COMMAND [<args>]""",
        prog="fal",
    )

    if len(argv) < 2:
        print("No command supplied\n")
        parser.print_help()
        exit(1)

    # Handle version checking
    version = pkg_resources.get_distribution("fal").version
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"fal {version}",
        help="show fal version",
    )

    # Handle commands
    command_parsers = parser.add_subparsers(title="commands", dest="command_parsers")
    run_parser = command_parsers.add_parser(
        "run",
        help="Run Python scripts as final nodes",
    )

    run_parser.add_argument(
        "--project-dir",
        default=os.getcwd(),
        help="Directory to look for dbt_project.yml.",
    )
    run_parser.add_argument(
        "--profiles-dir",
        default=None,
        help="Directory to look for profiles.yml.",
    )
    run_parser.add_argument(
        "--keyword",
        default="fal",
        help="Property in meta to look for fal configurations.",
    )

    run_parser.add_argument(
        "--all",
        action="store_true",
        help="Run scripts for all models. By default, fal runs scripts for models that ran in the last dbt run.",
    )
    # TODO: remove `action="extend"` to match exactly what dbt does
    run_parser.add_argument(
        "-s", "--select",
        nargs="+",
        action="extend", # For backwards compatibility with past fal version
        dest="select",
        help="Specify the nodes to include.",
    )
    run_parser.add_argument(
        "-m", "--models",
        nargs="+",
        action="extend", # For backwards compatibility with past fal version
        dest="select",
        help="Specify the nodes to include.",
    )
    run_parser.add_argument(
        "--exclude",
        nargs="+",
        action="extend", # For backwards compatibility with past fal version
        help="Specify the nodes to exclude.",
    )
    run_parser.add_argument(
        "--selector",
        help="The selector name to use, as defined in selectors.yml",
    )
    run_parser.add_argument(
        "--scripts",
        nargs="+",
        action="extend", # For backwards compatibility with past fal version
        help="Specify scripts to run, overrides schema.yml",
    )

    run_parser.add_argument(
        "--before",
        action="store_true",
        help="Run scripts specified in model `before` tag",
    )
    run_parser.add_argument(
        "--experimental-ordering",
        action="store_true",
        help="Turns on ordering of the fal scripts.",
    )
    run_parser.add_argument(
        "--debug",
        action="store_true",
        help="Display debug logging during execution.",
    )
    run_parser.add_argument(
        "--disable-logging",
        action="store_true",
        help="Disable logging.",
    )

    # fmt: on

    args = parser.parse_args(argv[1:])

    # TODO: remove `action="extend"` to match exactly what dbt does
    selects_count = (
        argv.count("-s")
        + argv.count("--select")
        + argv.count("-m")
        + argv.count("--model")
    )

    _run(
        project_dir=args.project_dir,
        profiles_dir=args.profiles_dir,
        keyword=args.keyword,
        all=args.all,
        select=args.select,
        exclude=args.exclude,
        selector=args.selector,
        script=args.scripts,
        before=args.before,
        experimental_ordering=args.experimental_ordering,
        debug=args.debug,
        disable_logging=args.disable_logging,
        selects_count=selects_count,
    )


def cli():
    run_fal(sys.argv)


def _run(
    project_dir,
    profiles_dir,
    keyword,
    all,
    select,
    exclude,
    selector,
    script,
    before,
    experimental_ordering,
    debug,
    disable_logging,
    # TODO: remove `action="extend"` to match exactly what dbt does
    selects_count,
):

    if disable_logging:
        logger.disable()

    # Re-enable logging for 1.0.0 through old API of logger
    elif DBT_VCURRENT.compare(DBT_V1) >= 0:
        if logger.disabled:
            logger.enable()

    with log_manager.applicationbound():
        if debug:
            log_manager.set_debug()

        real_project_dir = os.path.realpath(os.path.normpath(project_dir))
        real_profiles_dir = None
        if profiles_dir is not None:
            real_profiles_dir = os.path.realpath(os.path.normpath(profiles_dir))
        elif os.getenv("DBT_PROFILES_DIR"):
            real_profiles_dir = os.path.realpath(
                os.path.normpath(os.getenv("DBT_PROFILES_DIR"))
            )
        else:
            real_profiles_dir = DEFAULT_PROFILES_DIR

        selector_flags = select or exclude or selector
        if all and selector_flags:
            raise FalGeneralException(
                "Cannot pass --all flag alongside selection flags (--select/--models, --exclude, --selector)"
            )

        faldbt = FalDbt(
            real_project_dir, real_profiles_dir, select, exclude, selector, keyword
        )
        project = FalProject(faldbt)
        models = project.get_filtered_models(all, selector_flags, before)

        # TODO: remove `action="extend"` to match exactly what dbt does
        if selects_count > 1:
            dbt.exceptions.warn_or_error(
                "Passing multiple --select/--model flags to fal is deprecatd.\n"
                + f"Please use model selection like dbt. Use: --select {' '.join(select)}",
                log_fmt=dbt.ui.warning_tag("{}"),
            )

        print_run_info(models, keyword, before)

        if script:
            scripts = []
            for model in models:
                for el in script:
                    scripts.append(FalScript(model, el))
            return run_scripts(scripts, project)

        if experimental_ordering:
            scripts = ScriptGraph(models, keyword, project_dir).sort()
        else:
            scripts = []
            for model in models:
                for path in model.get_script_paths(keyword, real_project_dir, before):
                    scripts.append(FalScript(model, path))

        # run model specific scripts first
        run_scripts(scripts, project)

        # then run global scripts
        global_scripts = list(
            map(
                lambda path: FalScript(None, path, []),
                faldbt._global_script_paths,
            )
        )

        run_global_scripts(global_scripts, project)



from dataclasses import dataclass, field
from typing import List, TypeVar, Dict
from faldbt.project import DbtModel, FalDbt
from pathlib import Path
import re

T = TypeVar("T", bound="FalScript")


class FalDagCycle(Exception):
    pass


@dataclass(frozen=True)
class FalScript:
    model: DbtModel
    path: str
    dependencies: List[T] = field(default_factory=list)

    def __hash__(self):
        return self.path.__hash__() * self.model.name.__hash__()

    def __repr__(self):
        return "FalScript(" + self.path.__str__() + ")"

    def parse(self) -> List[str]:
        """
        Parses the python file to identify any dependencies
        """
        with open(self.path) as file:
            lines = file.readlines()
            ## TODO: do this search in parallel
            dependency_names = []
            for line in lines:
                match = re.search('ref\("(.*)"\)', line)
                if match:
                    group = match.group(1)
                    dependency_names.append(group)
            return dependency_names

    def add_dependency(self, dependency: T):
        if dependency not in self.dependencies and dependency != self:
            self.dependencies.append(dependency)

    def exec(self, context, faldbt: FalDbt):
        """
        Executes the script
        """
        with open(self.path) as file:
            a_script = file.read()
            exec(
                a_script,
                {
                    "context": context,
                    "ref": faldbt.ref,
                    "source": faldbt.source,
                    "write_to_source": faldbt.write_to_source,
                    "write_to_firestore": faldbt.write_to_firestore,
                    "list_models": faldbt.list_models,
                    "list_models_ids": faldbt.list_models_ids,
                    "list_sources": faldbt.list_sources,
                    "list_features": faldbt.list_features
                },
            )


@dataclass(frozen=True)
class _ScriptUniqueKey:
    model: DbtModel
    script_path: Path


class _ScriptGraphBuilder:
    modelNameToModelLookup: Dict[str, DbtModel] = {}
    modelToScriptLookup: Dict[str, List[Path]] = {}
    falScripts: Dict[_ScriptUniqueKey, FalScript] = {}

    def __init__(self, models: List[DbtModel], keyword: str, project_dir: str, before: bool = False):
        for model in models:
            self.modelToScriptLookup[model.name] = model.get_script_paths(
                keyword, project_dir, before=before
            )
            self.modelNameToModelLookup[model.name] = model
        for model_name, script_list in self.modelToScriptLookup.items():
            for script in script_list:
                self.recursively_set_dependencies(
                    _ScriptUniqueKey(self.modelNameToModelLookup[model_name], script)
                )

    def recursively_set_dependencies(self, key: _ScriptUniqueKey):
        if key in self.falScripts:
            return
        self.falScripts[key] = FalScript(key.model, key.script_path)
        current_script = self.falScripts.get(key)
        dependency_models = current_script.parse()

        for model_name in dependency_models:
            dependency_model = self.modelNameToModelLookup[model_name]
            dependency_scripts = self.modelToScriptLookup.get(model_name, [])

            for dependency_script in dependency_scripts:
                dependency_key = _ScriptUniqueKey(dependency_model, dependency_script)
                self.recursively_set_dependencies(dependency_key)
                current_script.add_dependency(self.falScripts.get(dependency_key))

    def get_values(self) -> List[FalScript]:
        return self.falScripts.values()


class ScriptGraph:
    def __init__(
        self,
        models: List[DbtModel],
        keyword: str,
        project_dir: str,
        _graph: List[FalScript] = [],
    ):
        self.outgoing: Dict[FalScript, List[FalScript]] = {}
        self.incoming: Dict[FalScript, List[FalScript]] = {}
        self.ordered_list: List[FalScript] = []

        if _graph:
            self.graph = _graph
        else:
            self.graph = _ScriptGraphBuilder(models, keyword, project_dir).get_values()
        self.incoming = dict(
            map(lambda script: [script, script.dependencies], self.graph)
        )
        for edge in self.graph:
            self.outgoing[edge] = []
        for edge in self.graph:
            for dependency in edge.dependencies:
                self.outgoing[dependency].append(edge)

    def sort(self):
        """
        Topologically sorts the python scripts to make sure all dependencies
        run before the script itsef
        """
        leaf_nodes = list(
            filter(lambda key: len(self.incoming[key]) == 0, self.incoming.keys())
        )
        while leaf_nodes:
            node = leaf_nodes.pop(0)
            self.ordered_list.append(node)
            outgoing_copy = self.outgoing[node].copy()
            for item in outgoing_copy:
                self.incoming[item].remove(node)
                self.outgoing[node].remove(item)
                if not self.incoming[item]:
                    leaf_nodes.append(item)
        detect_cycles = _flatten(self.outgoing.values())
        if detect_cycles:
            raise FalDagCycle(
                "Your python scripts contain a cycle could not determine the right order"
            )
        return self.ordered_list


def _flatten(forest):
    return [leaf for tree in forest for leaf in tree]



# NOTE: COPIED FROM https://github.com/dbt-labs/dbt-core/blob/89907f09c8a94d95d2882f909f05143e098746a0/core/dbt/contracts/sql.py
import uuid
from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional, List, Any, Dict, Sequence

from dbt.dataclass_schema import dbtClassMixin

from dbt.contracts.graph.compiled import CompileResultNode
from dbt.contracts.results import (
    RunResult, RunResultsArtifact, TimingInfo,
    ExecutionResult,
    RunExecutionResult,
)
from dbt.contracts.util import VersionedSchema, schema_version
from dbt.logger import LogMessage


TaskTags = Optional[Dict[str, Any]]
TaskID = uuid.UUID

# Outputs


@dataclass
class RemoteResult(VersionedSchema):
    logs: List[LogMessage]


@dataclass
class RemoteCompileResultMixin(RemoteResult):
    raw_sql: str
    compiled_sql: str
    node: CompileResultNode
    timing: List[TimingInfo]


@dataclass
@schema_version('remote-compile-result', 1)
class RemoteCompileResult(RemoteCompileResultMixin):
    generated_at: datetime = field(default_factory=datetime.utcnow)

    @property
    def error(self):
        return None


@dataclass
@schema_version('remote-execution-result', 1)
class RemoteExecutionResult(ExecutionResult, RemoteResult):
    results: Sequence[RunResult]
    args: Dict[str, Any] = field(default_factory=dict)
    generated_at: datetime = field(default_factory=datetime.utcnow)

    def write(self, path: str):
        writable = RunResultsArtifact.from_execution_results(
            generated_at=self.generated_at,
            results=self.results,
            elapsed_time=self.elapsed_time,
            args=self.args,
        )
        writable.write(path)

    @classmethod
    def from_local_result(
        cls,
        base: RunExecutionResult,
        logs: List[LogMessage],
    ) -> 'RemoteExecutionResult':
        return cls(
            generated_at=base.generated_at,
            results=base.results,
            elapsed_time=base.elapsed_time,
            args=base.args,
            logs=logs,
        )


@dataclass
class ResultTable(dbtClassMixin):
    column_names: List[str]
    rows: List[Any]


@dataclass
@schema_version('remote-run-result', 1)
class RemoteRunResult(RemoteCompileResultMixin):
    table: ResultTable
    generated_at: datetime = field(default_factory=datetime.utcnow)



# NOTE: INSPIRED IN https://github.com/dbt-labs/dbt-core/blob/43edc887f97e359b02b6317a9f91898d3d66652b/core/dbt/lib.py
import six
from datetime import datetime
from dataclasses import dataclass
from uuid import uuid4
from typing import List, Tuple, Union

import dbt.version
import dbt.semver
import dbt.flags as flags
import dbt.adapters.factory as adapters_factory
from dbt.config.runtime import RuntimeConfig
from dbt.contracts.connection import AdapterResponse
from dbt.contracts.graph.manifest import Manifest
from dbt.parser.manifest import process_node
from dbt.logger import GLOBAL_LOGGER as logger

from . import parse

import pandas as pd
from pandas.io import sql as pdsql

import sqlalchemy
from sqlalchemy.sql.ddl import CreateTable
from sqlalchemy.sql import Insert
from sqlalchemy.sql.schema import MetaData


DBT_V1 = dbt.semver.VersionSpecifier.from_version_string("1.0.0")
DBT_VCURRENT = dbt.version.get_installed_version()

if DBT_VCURRENT.compare(DBT_V1) >= 0:
    from dbt.parser.sql import SqlBlockParser
    from dbt.contracts.graph.parsed import ParsedModelNode, ParsedSourceDefinition
    from dbt.contracts.sql import ResultTable, RemoteRunResult
else:
    from faldbt.cp.parser.sql import SqlBlockParser
    from faldbt.cp.contracts.graph.parsed import ParsedModelNode, ParsedSourceDefinition
    from faldbt.cp.contracts.sql import ResultTable, RemoteRunResult


@dataclass
class FlagsArgs:
    profiles_dir: str
    use_colors: bool


def initialize_dbt_flags(profiles_dir: str):
    """
    Initializes the flags module from dbt, since it's accessed from around their code.
    """
    args = FlagsArgs(profiles_dir, None)
    user_config = parse.get_dbt_user_config(profiles_dir)
    try:
        flags.set_from_args(args, user_config)
    except TypeError:
        flags.set_from_args(args)

    # Set invocation id
    if DBT_VCURRENT.compare(DBT_V1) >= 0:
        import dbt.events.functions as events_functions

        events_functions.set_invocation_id()

    # Re-enable logging for 1.0.0 through old API of logger
    # TODO: migrate for 1.0.0 code to new event system
    if DBT_VCURRENT.compare(DBT_V1) >= 0:
        flags.ENABLE_LEGACY_LOGGER = "1"


def register_adapters(config: RuntimeConfig):
    # Clear previously registered adapters. This fixes cacheing behavior on the dbt-server
    adapters_factory.reset_adapters()
    # Load the relevant adapter
    adapters_factory.register_adapter(config)


def _get_operation_node(manifest: Manifest, project_path, profiles_dir, sql):

    config = parse.get_dbt_config(project_path, profiles_dir)
    block_parser = SqlBlockParser(
        project=config,
        manifest=manifest,
        root_project=config,
    )

    # NOTE: nodes get registered to the manifest automatically,
    # HACK: we need to include uniqueness (UUID4) to avoid clashes
    name = "SQL:" + str(hash(sql)) + ":" + str(uuid4())
    sql_node = block_parser.parse_remote(sql, name)
    process_node(config, manifest, sql_node)
    return sql_node


# NOTE: Once we get an adapter, we must call `connection_for` or `connection_named` to use it
def _get_adapter(project_path: str, profiles_dir: str):
    config = parse.get_dbt_config(project_path, profiles_dir)

    adapters_factory.cleanup_connections()
    return adapters_factory.get_adapter(config)


def _execute_sql(
    manifest: Manifest, project_path: str, profiles_dir: str, sql: str
) -> Tuple[AdapterResponse, RemoteRunResult]:
    node = _get_operation_node(manifest, project_path, profiles_dir, sql)
    adapter = _get_adapter(project_path, profiles_dir)

    logger.debug("Running query\n{}", sql)

    result = None
    with adapter.connection_for(node):
        adapter.connections.begin()
        response, execute_result = adapter.execute(sql, fetch=True)

        table = ResultTable(
            column_names=list(execute_result.column_names),
            rows=[list(row) for row in execute_result],
        )

        result = RemoteRunResult(
            raw_sql=sql,
            compiled_sql=sql,
            node=node,
            table=table,
            timing=[],
            logs=[],
            generated_at=datetime.utcnow(),
        )
        adapter.connections.commit()

    return response, result


def _get_target_relation(
    target: Union[ParsedModelNode, ParsedSourceDefinition],
    project_path: str,
    profiles_dir: str,
):
    adapter = _get_adapter(project_path, profiles_dir)

    relation = None
    with adapter.connection_named(str(uuid4())):
        # This ROLLBACKs so it has to be a new connection
        relation = adapter.get_relation(
            target.database, target.schema, target.identifier
        )
    return relation


def execute_sql(
    manifest: Manifest, project_path: str, profiles_dir: str, sql: str
) -> RemoteRunResult:
    _, result = _execute_sql(manifest, project_path, profiles_dir, sql)
    return result


def fetch_target(
    manifest: Manifest,
    project_path: str,
    profiles_dir: str,
    target: Union[ParsedModelNode, ParsedSourceDefinition],
) -> RemoteRunResult:
    relation = _get_target_relation(target, project_path, profiles_dir)

    if relation is None:
        raise Exception(f"Could not get relation for '{target.unique_id}'")

    query = f"SELECT * FROM {relation}"
    _, result = _execute_sql(manifest, project_path, profiles_dir, query)
    return result


def write_target(
    data: pd.DataFrame,
    manifest: Manifest,
    project_path: str,
    profiles_dir: str,
    target: Union[ParsedModelNode, ParsedSourceDefinition],
) -> RemoteRunResult:
    adapter = _get_adapter(project_path, profiles_dir)

    relation = _get_target_relation(target, project_path, profiles_dir)

    engine = _alchemy_engine(adapter, target)
    pddb = pdsql.SQLDatabase(engine, schema=target.schema)
    pdtable = pdsql.SQLTable(target.name, pddb, data, index=False)
    alchemy_table: sqlalchemy.Table = pdtable.table.to_metadata(pdtable.pd_sql.meta)

    column_names: List[str] = list(data.columns)
    rows = data.to_records(index=False)
    row_dicts = list(map(lambda row: dict(zip(column_names, row)), rows))

    if relation is None:
        create_stmt = CreateTable(alchemy_table).compile(
            bind=engine, compile_kwargs={"literal_binds": True}
        )

        _execute_sql(
            manifest, project_path, profiles_dir, six.text_type(create_stmt).strip()
        )

    insert_stmt = (
        Insert(alchemy_table)
        .values(row_dicts)
        .compile(bind=engine, compile_kwargs={"literal_binds": True})
    )

    _, result = _execute_sql(
        manifest, project_path, profiles_dir, six.text_type(insert_stmt).strip()
    )
    return result


def _alchemy_engine(
    adapter: adapters_factory.Adapter,
    target: Union[ParsedModelNode, ParsedSourceDefinition],
):
    url_string = f"{adapter.type()}://"
    if adapter.type() == "bigquery":
        url_string = f"bigquery://{target.database}"
    if adapter.type() == "postgres":
        url_string = "postgresql://"
    # TODO: add special cases as needed

    def null_dump(sql, *multiparams, **params):
        pass

    return sqlalchemy.create_mock_engine(url_string, executor=null_dump)



# NOTE: COPIED FROM https://github.com/dbt-labs/dbt-core/blob/89907f09c8a94d95d2882f909f05143e098746a0/core/dbt/parser/sql.py
import os
from dataclasses import dataclass
from typing import Iterable

from dbt.contracts.graph.manifest import SourceFile
from dbt.contracts.graph.unparsed import UnparsedMacro
from dbt.exceptions import InternalException
from dbt.parser.base import SimpleSQLParser
from dbt.parser.macros import MacroParser
from dbt.parser.search import FileBlock

from faldbt.cp.contracts.graph.parsed import ParsedSqlNode, ParsedMacro
from faldbt.cp.node_types import NodeType


@dataclass
class SqlBlock(FileBlock):
    block_name: str

    @property
    def name(self):
        return self.block_name


class SqlBlockParser(SimpleSQLParser[ParsedSqlNode]):
    def parse_from_dict(self, dct, validate=True) -> ParsedSqlNode:
        if validate:
            ParsedSqlNode.validate(dct)
        return ParsedSqlNode.from_dict(dct)

    @property
    def resource_type(self) -> NodeType:
        return NodeType.SqlOperation

    @staticmethod
    def get_compiled_path(block: FileBlock):
        # we do it this way to make mypy happy
        if not isinstance(block, SqlBlock):
            raise InternalException(
                'While parsing SQL operation, got an actual file block instead of '
                'an SQL block: {}'.format(block)
            )

        return os.path.join('sql', block.name)

    def parse_remote(self, sql: str, name: str) -> ParsedSqlNode:
        source_file = SourceFile.remote(sql, self.project.project_name)
        contents = SqlBlock(block_name=name, file=source_file)
        return self.parse_node(contents)


class SqlMacroParser(MacroParser):
    def parse_remote(self, contents) -> Iterable[ParsedMacro]:
        base = UnparsedMacro(
            path='from remote system',
            original_file_path='from remote system',
            package_name=self.project.project_name,
            raw_sql=contents,
            root_path=self.project.project_root,
            resource_type=NodeType.Macro,
        )
        for node in self.parse_unparsed_macros(base):
            yield node



"""Run fal scripts."""
import os
from typing import Dict, Any, List, Union
from dataclasses import dataclass
from dbt.config.runtime import RuntimeConfig
from pathlib import Path

from dbt.contracts.results import NodeStatus
from dbt.logger import GLOBAL_LOGGER as logger

from faldbt.project import FalProject
from fal.dag import FalScript

import faldbt.lib as lib

if lib.DBT_VCURRENT.compare(lib.DBT_V1) >= 0:
    from dbt.contracts.graph.parsed import ColumnInfo
else:
    from faldbt.cp.contracts.graph.parsed import ColumnInfo


@dataclass
class CurrentModel:
    name: str
    status: NodeStatus
    columns: Dict[str, ColumnInfo]
    tests: List[Any]
    meta: Dict[Any, Any]


@dataclass
class CurrentTest:
    name: str
    modelname: str
    column: str
    status: str


@dataclass
class ContextConfig:
    target_path: Path


@dataclass
class Context:
    current_model: Union[CurrentModel, None]
    config: ContextConfig


def run_scripts(list: List[FalScript], project: FalProject):
    faldbt = project._faldbt
    for script in list:
        model = script.model
        meta = model.meta
        _del_key(meta, project.keyword)

        tests = _process_tests(model.tests)

        current_model = CurrentModel(
            name=model.name,
            status=model.status,
            columns=model.columns,
            tests=tests,
            meta=meta,
        )

        context_config = ContextConfig(_get_target_path(faldbt._config))
        context = Context(current_model=current_model, config=context_config)

        logger.info("Running script {} for model {}", script.path, model.name)

        script.exec(context, faldbt)


def run_global_scripts(list: List[FalScript], project: FalProject):
    faldbt = project._faldbt
    for script in list:
        context_config = ContextConfig(_get_target_path(faldbt._config))
        context = Context(current_model=None, config=context_config)

        logger.info("Running global script {}", script.path)

        script.exec(context, faldbt)


def _del_key(dict: Dict[str, Any], key: str):
    try:
        del dict[key]
    except KeyError:
        pass


def _get_target_path(config: RuntimeConfig) -> Path:
    return Path(os.path.realpath(os.path.join(config.project_root, config.target_path)))


def _process_tests(tests: List[Any]):
    return list(map(
        lambda test: CurrentTest(name=test.name,
                                 column=test.column,
                                 status=test.status,
                                 modelname=test.model), tests))



# NOTE: COPIED FROM https://github.com/dbt-labs/dbt-core/blob/89907f09c8a94d95d2882f909f05143e098746a0/core/dbt/node_types.py
from typing import List

from dbt.dataclass_schema import StrEnum


class NodeType(StrEnum):
    Model = 'model'
    Analysis = 'analysis'
    Test = 'test'
    Snapshot = 'snapshot'
    Operation = 'operation'
    Seed = 'seed'
    # TODO: rm?
    RPCCall = 'rpc'
    SqlOperation = 'sql'
    Documentation = 'docs'
    Source = 'source'
    Macro = 'macro'
    Exposure = 'exposure'

    @classmethod
    def executable(cls) -> List['NodeType']:
        return [
            cls.Model,
            cls.Test,
            cls.Snapshot,
            cls.Analysis,
            cls.Operation,
            cls.Seed,
            cls.Documentation,
            cls.RPCCall,
            cls.SqlOperation
        ]

    @classmethod
    def refable(cls) -> List['NodeType']:
        return [
            cls.Model,
            cls.Seed,
            cls.Snapshot,
        ]

    @classmethod
    def documentable(cls) -> List['NodeType']:
        return [
            cls.Model,
            cls.Seed,
            cls.Snapshot,
            cls.Source,
            cls.Macro,
            cls.Analysis,
            cls.Exposure
        ]

    def pluralize(self) -> str:
        if self == 'analysis':
            return 'analyses'
        else:
            return f'{self}s'


class RunHookType(StrEnum):
    Start = 'on-run-start'
    End = 'on-run-end'
